{
  glSOAR

  ObjectPascal implementation of SOAR terrain rendering algorithm using OpenGL
  by Marek Mauder (pentar@seznam.cz)

  The contents of this file are used with permission, subject to the Mozilla
  Public License Version 1.1 (the "License"); you may not use this file except
  in compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/MPL-1.1.html

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
  the specific language governing rights and limitations under the License.

  Alternatively, the contents of this file may be used under the terms of the
  GNU Lesser General Public License (the  "LGPL License"), in which case the
  provisions of the LGPL License are applicable instead of those above.
  If you wish to allow use of your version of this file only under the terms
  of the LGPL License and not to allow others to use your version of this file
  under the MPL, indicate your decision by deleting  the provisions above and
  replace  them with the notice and other provisions required by the LGPL
  License.  If you do not delete the provisions above, a recipient may use
  your version of this file under either the MPL or the LGPL License.

  For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html
}

program glSOAR;

{$APPTYPE CONSOLE}

uses
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}  
  SysUtils,
  gl,
  glu,
  glut,
  glext,
  glSOARHelpers,
  glSOARMain,
  VUMaths,
  ImagingUtility;

const
  SWinTitle  = 'glSOAR';
  STimeStamp = '11th May 2006';

  SDataDir = 'Data';
  SDefaultHeightmap     = SDataDir + PathDelim + 'Volcanoes.png';
  SDefaultGround        = SDataDir + PathDelim + 'VolcanoesTX.dds';
  SDefaultPath          = SDataDir + PathDelim + 'Volcanoes.pt';
  S9x9Path              = SDataDir + PathDelim + 'Volcanoes9x9.pt';
  SMountainousHeightmap = SDataDir + PathDelim + 'Everest.png';
  SMountainousGround    = SDataDir + PathDelim + 'EverestTX.dds';
  SMountainousPath      = SDataDir + PathDelim + 'Everest.pt';
  SFlatHeightmap        = SDataDir + PathDelim + 'Puget.png';
  SFlatGround           = SDataDir + PathDelim + 'PugetTX.dds';
  SFlatPath             = SDataDir + PathDelim + 'Puget.pt';
  SDetailTexFile        = SDataDir + PathDelim + 'Detail.dds';

  WalkSpeed  = 1.8;
  RunSpeed   = 7.0;
  UltraSpeed = 50.0;

  DefaultTerrainSize  = 1025;
  DefaultScreenWidth  = 800;
  DefaultScreenHeight = 600;

var
  // Program init variables
  ScreenWidth: LongInt = DefaultScreenWidth;
  ScreenHeight: LongInt = DefaultScreenHeight;
  FullScreen: Boolean = False;
  TerrainSize: LongInt = DefaultTerrainSize;
  TerrainZScale: Single = 0.25;
  TerrainBlocks: LongInt = 1;
  Filter8BitHeightmaps: Boolean = True;
  StartX: LongInt = 0;
  StartY: LongInt = 0;

  // User changeable variables
  ActiveStats: Boolean = False;
  ActiveWireOverlay: Boolean = False;
  ActiveFlying: Boolean = False;
  ActiveTexturing: Boolean = True;
  ActiveFog: Boolean = False;
  ActiveDetailTexture: Boolean = True;
  CullFovX: Single = 45.0;

  // Benchmark variables
  BenchmarkMode: Boolean = False;
  BenchmarkStage: LongInt = 0;
  BenchmarkStageName: string;
  BenchmarkStageCount: LongInt = 11;
  BenchmarkTime: Single;
  BenchmarkLength: Single;
  CameraAnimation: TAnimation;
  BenchmarkFrames: LongInt;
  BenchmarkTriangles: Int64;
  BenchmarkMinFPS: Single;
  BenchmarkMaxFPS: Single;
  BenchmarkMinTPF: Single;
  BenchmarkMaxTPF: Single;
  BenchmarkRenderTime: Int64;
  BenchmarkRefineTime: Int64;

  // Camera variables
  Position: TVector3 = (X: 0.0; Y: 0.0; Z: 0.0);
  Direction: TVector3 = (X: 1.0; Y: 1.0; Z: 0.0);
  ViewUp: TVector3;
  ViewAt: TVector3;
  ViewEye: TVector3;
  Frustum: TFrustum;
  DrawPath: Boolean = False;

var
  // Private variables
  LastTime: Int64 = 0;
  FrameTime: Int64 = 0;
  RefineTime, RenderTime: Int64;
  Frames: LongInt = 0;
  FPS, TPS: Single;
  FovX: Single = 45.0;
  NearZ: Single = 0.1;
  FarZ: Single = 131072.0;
  SkyColor: TVector4 = (X: 0.5; Y: 0.625; Z: 0.75; W: 1.0);
  DetailRepeat: Single = 2.0;
  GroundColor: TVector4 = (X: 0.60; Y: 0.45; Z: 0.30; W: 1.0);
  TrianglesPerFrame, TrianglesDrawn, TotalTerrainTriangles: LongInt;
  DrawFPS: Boolean = True;
  HeightmapFile: string;
  GroundTexFile: string;
  GroundTexture: GLuint = 0;
  DetailTexture: GLuint = 0;

{ Application procedures }

procedure CheckInput(ElapsedTime: Single); forward;
procedure RunBenchmark(UpdateOnly: Boolean; const ElapsedTime: Single); forward;

procedure InitAppData;
var
  TexGenSPlane, TexGenTPlane: TVector4;
begin
  // Set default GL settings 
  glHint(GL_FOG_HINT, GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

  glDisable(GL_LIGHTING);
  glDisable(GL_BLEND);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_TEXTURE_2D);

  glFrontFace(GL_CW);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  if ActiveFog then
    glEnable(GL_FOG);

  glFogi(GL_FOG_MODE, GL_EXP2);
  glFogf(GL_FOG_DENSITY, 1.0 / (TerrainSize * Sqrt(TerrainBlocks)));
  glFogfv(GL_FOG_COLOR, @SkyColor);

  glEnableClientState(GL_VERTEX_ARRAY);

  // Load terrain and textures
  MsgOut('');
  MsgOut(Format('Loading terrain (size: %dx%d, blocks: %dx%d)', [TerrainSize,
    TerrainSize, TerrainBlocks, TerrainBlocks]));
  LoadTerrain(HeightmapFile, TerrainSize, TerrainZScale, TerrainBlocks > 1, Filter8BitHeightmaps);
  MsgOut(Format('Heightmap loaded: %s', [ExtractFileName(HeightmapFile)]));
  GroundTexture := LoadTexture(GroundTexFile);
  DetailTexture := LoadTexture(GetAppDir + PathDelim + SDetailTexFile);
  MsgOut('Terrain loaded successfully');
  TotalTerrainTriangles := TerrainSize * TerrainSize * 2 * TerrainBlocks * TerrainBlocks;

  // Set ground texture
  glBindTexture(GL_TEXTURE_2D, GroundTexture);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  TexGenSPlane := Vector4(1.0 / TerrainSize, 0.0, 0.0, 0.0);
  TexGenTPlane := Vector4(0.0, 1.0 / TerrainSize, 0.0, 0.0);
  glTexGenfv(GL_S, GL_OBJECT_PLANE, @TexGenSPlane);
	glTexGenfv(GL_T, GL_OBJECT_PLANE, @TexGenTPlane);

  // Set detail texture if multitexturing is supported
  MultiTexturing := Load_GL_ARB_multitexture;
  if MultiTexturing then
  begin
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_2D, DetailTexture);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
    glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 2);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    TexGenSPlane := Vector4(1.0 / DetailRepeat, 0.0, 0.0, 0.0);
    TexGenTPlane := Vector4(0.0, 1.0  / DetailRepeat, 0.0, 0.0);
    glTexGenfv(GL_S, GL_OBJECT_PLANE, @TexGenSPlane);
    glTexGenfv(GL_T, GL_OBJECT_PLANE, @TexGenTPlane);
    glActiveTextureARB(GL_TEXTURE0_ARB);
  end;

  LastTime := GetTimeMicroseconds;
  FrameTime := LastTime;
  RefineTime := 0;
  RenderTime := 0;
end;

procedure FreeAppData;
begin
  FreeTerrain;
  glDeleteTextures(1, @GroundTexture);
  glDeleteTextures(2, @DetailTexture);
end;

procedure TerminateApp;
begin
  FreeAppData;
  Halt(0);
end;

procedure ResizeWindow(NewWidth, NewHeight: LongInt);
begin
  glViewport(0, 0, NewWidth, NewHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(FovX, NewWidth / NewHeight, NearZ, FarZ);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  ScreenWidth := NewWidth;
  ScreenHeight := NewHeight;
  ReturnMouse;
  glutShowWindow;
end;

procedure DrawTerrainBlock(X, Y: LongInt; var RefineTime, RenderTime: Int64);
var
  Time: Int64;
  ShiftVector: TVector3;
  EdgeFlags: TEdgeFlags;
begin
  ShiftVector := Vector3(X * (TerrainSize - 1), Y * (TerrainSize - 1), 0.0);
  EdgeFlags := [];

  if (X >= 0) and (X < TerrainBlocks - 1) then Include(EdgeFlags, efPosX);
  if (X > 0) and (X <= TerrainBlocks - 1) then Include(EdgeFlags, efNegX);
  if (Y >= 0) and (Y < TerrainBlocks - 1) then Include(EdgeFlags, efPosY);
  if (Y > 0) and (Y <= TerrainBlocks - 1) then Include(EdgeFlags, efNegY);

  Time := GetTimeMicroseconds;
  MeshRefine(ScreenWidth, FovX, EdgeFlags, ViewEye, ShiftVector, Frustum);
  Inc(RefineTime, GetTimeMicroseconds - Time);

  // Enable/disable textures
  if ActiveTexturing then
  begin
    glEnable(GL_TEXTURE_2D);
    if MultiTexturing then
    begin
      glActiveTextureARB(GL_TEXTURE1_ARB);
      if ActiveDetailTexture then
        glEnable(GL_TEXTURE_2D)
      else
        glDisable(GL_TEXTURE_2D);
      glActiveTextureARB(GL_TEXTURE0_ARB);
    end;
  end
  else
  begin
    DisableTextures;
    glColor4fv(@GroundColor);
  end;  

  if ActiveWireOverlay then
  begin
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 4.0);
  end;

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix;
  glTranslatef(ShiftVector.X, ShiftVector.Y, ShiftVector.Z);

  glVertexPointer(3, GL_FLOAT, SizeOf(TVertex), @VertexBuffer[0].Pos);

  Time := GetTimeMicroseconds;
  // Draw terrain block
  glDrawElements(GL_TRIANGLE_STRIP, NumIndices, GL_UNSIGNED_INT, IndexBuffer);
  // Draw block edges if necessary
  if efPosX in EdgeFlags then
    glDrawElements(GL_TRIANGLES, NumPosXEdgeIndices, GL_UNSIGNED_INT, PosXEdgeIndices);
  if efPosY in EdgeFlags then
    glDrawElements(GL_TRIANGLES, NumPosYEdgeIndices, GL_UNSIGNED_INT, PosYEdgeIndices);
  if efNegX in EdgeFlags then
    glDrawElements(GL_TRIANGLES, NumNegXEdgeIndices, GL_UNSIGNED_INT, NegXEdgeIndices);
  if efNegY in EdgeFlags then
    glDrawElements(GL_TRIANGLES, NumNegYEdgeIndices, GL_UNSIGNED_INT, NegYEdgeIndices);
  Inc(RenderTime, GetTimeMicroseconds - Time);

  if ActiveWireOverlay then
  begin
    glPushAttrib(GL_ENABLE_BIT or GL_COLOR_BUFFER_BIT or GL_POLYGON_BIT);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glEnable(GL_LINE_SMOOTH);
    DisableTextures;
    glDrawElements(GL_TRIANGLE_STRIP, NumIndices, GL_UNSIGNED_INT, IndexBuffer);
    glPopAttrib;
    glDisable(GL_POLYGON_OFFSET_FILL);
  end;

  glPopMatrix;

  Inc(TrianglesDrawn, NumTriangles);
end;

procedure DrawScene;
var
  X, Y, I: LongInt;
  ElapsedTime: Single;
  CurrTime: Int64;
  OldTrianglesDrawn: LongInt;
begin
  OldTrianglesDrawn := TrianglesDrawn;

  CurrTime := GetTimeMicroseconds;
  ElapsedTime := (CurrTime - LastTime) / 1e06;
  LastTime := CurrTime;

  if BenchmarkMode then
    RunBenchmark(True, ElapsedTime)
  else
    CheckInput(ElapsedTime);

  // Update camera vectors
  ViewEye := Vector3(Position.X, Position.Y, Position.Z + 1.60);
  ViewAt := Vector3(Position.X + Direction.X, Position.Y + Direction.Y,
    Position.Z + Direction.Z + 1.60);
  ViewUp := Vector3(0.0, 0.0, 1.0);

  // Set camera position and target point
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(ViewEye.X, ViewEye.Y, ViewEye.Z,
    ViewAt.X, ViewAt.Y, ViewAt.Z,
    ViewUp.X, ViewUp.Y, ViewUp.Z);

  // Update frustum for view culling
  glMatrixMode(GL_PROJECTION);
  glPushMatrix;
  glLoadIdentity;
  gluPerspective(CullFovX, ScreenWidth / ScreenHeight, NearZ, FarZ);
  UpdateFrustum(Frustum);
  glPopMatrix;
  glMatrixMode(GL_MODELVIEW);

  glClearColor(SkyColor.X, SkyColor.Y, SkyColor.Z, 1.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  RefineTime := 0;
  RenderTime := 0;

  // Draw individual terrain blocks
  for X := 0 to TerrainBlocks - 1 do
    for Y := 0 to TerrainBlocks - 1 do
      DrawTerrainBlock(X, Y, RefineTime, RenderTime);

  if DrawPath then
  begin
    glPushAttrib(GL_LINE_BIT or GL_CURRENT_BIT or GL_ENABLE_BIT or GL_COLOR_BUFFER_BIT);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glEnable(GL_LINE_SMOOTH);
    DisableTextures;
    glLineWidth(8.0);
    glColor3f(1.0, 0.0, 0.0);

    glBegin(GL_LINE_STRIP);
    for I := 0 to CameraAnimation.FrameCount - 1 do
    begin
      glVertex3fv(@CameraAnimation.Frames[I].Position);
    end;
    glEnd;

    glPopAttrib;
  end;

  Inc(Frames);
  TrianglesPerFrame := TrianglesDrawn - OldTrianglesDrawn;

  if (CurrTime - FrameTime) > 1e06 then
  begin
    FPS := Frames / (CurrTime - FrameTime) * 1e06;
    Frames := 0;
    FrameTime := CurrTime;
    TPS := TrianglesDrawn;
    TrianglesDrawn := 0;
  end;

  if DrawFPS or ActiveStats then
  begin
    BeginText;
    if DrawFPS then
    begin
      TextOut(5, 0,  Format('FPS: %.0n', [FPS]), $FFFFFFC0);
      TextOut(5, 15, Format('TPS: %.0n', [TPS]), $FFFFFFC0);
    end;
    if ActiveStats then
    begin
      TextOut(5, 30, Format('Tolerance: %d pixels', [TolerancePixels]), $FFFFFFC0);
      TextOut(5, 45, Format('Camera FOV: %.0n  Culling FOV: %.0n', [FovX, CullFovX]), $FFFFFFC0);
      TextOut(5, 60, Format('Refinement Min LOD: %d / %d', [MinLOD, RefinementLevels]), $FFFFFFC0);
      TextOut(5, 75, Format('Refinement Time: %.0n us', [RefineTime * 1.0]), $FFFFFFC0);
      TextOut(5, 90, Format('Render Time: %.0n us', [RenderTime * 1.0]), $FFFFFFC0);
      TextOut(5, 105, Format('Refinement Efficiency: %.0n / %.0n (%.3n%%) triangles',
        [TrianglesPerFrame * 1.0, TotalTerrainTriangles * 1.0,
        (TrianglesPerFrame / TotalTerrainTriangles) * 100.0]), $FFFFFFC0);
      TextOut(5, 120, Format('Error Metric: %s', [Iff(ErrorMetric = emAnisotropic,
        'Anisotropic', 'Isotropic')]), $FFFFFFC0);
      TextOut(5, 135, Format('Camera Position: X: %.2n  Y: %.2n  Z: %.2n ', [ViewEye.X,
        ViewEye.Y, ViewEye.Z]), $FFFFFFC0);
      TextOut(5, 150, Format('Camera Direction: X: %.2n  Y: %.2n  Z: %.2n ', [Direction.X,
        Direction.Y, Direction.Z]), $FFFFFFC0);
    end;
    EndText;
  end;

  glFlush;
  glutSwapBuffers;
end;

procedure EndBenchmark;
begin
  TerminateApp;
end;

procedure StoreResults(Stage: LongInt);
var
  AvgFPS, AvgTPF, AvgEfficiency: Single;
  PathFPS, PathTPF, PathTimes: string;
  I: LongInt;
begin
  AvgFPS := BenchmarkFrames / BenchmarkLength;
  AvgTPF := BenchmarkTriangles / BenchmarkFrames;
  AvgEfficiency := AvgTPF / TotalTerrainTriangles * 100.0;

  PathFPS := '';
  PathTPF := '';
  PathTimes := '';
  for I := 0 to CameraAnimation.FrameCount - 1 do
  begin
    PathFPS := PathFPS + Format(' %4.0n ', [CameraAnimation.Frames[I].FPS]);
    PathTPF := PathTPF + Format(' %7.0n ', [CameraAnimation.Frames[I].TPF]);
    PathTimes := PathTimes + Format(' %3.1n ', [CameraAnimation.Frames[I].Time]);
  end;

  MsgOut('');
  MsgOut(Format('Benchmark Stage: %d', [Stage]));
  MsgOut(Format('  Stage Name: %s', [BenchmarkStageName]));
  MsgOut(Format('  Terrain Settings:', []));
  MsgOut(Format('    Heightmap:         %s', [ExtractFileName(HeightmapFile)]));
  MsgOut(Format('    Ground Tex:        %s', [ExtractFileName(GroundTexFile)]));
  MsgOut(Format('    Terrain Size:      %dx%d', [TerrainSize, TerrainSize]));
  MsgOut(Format('    Terrain Tri Count: %.0n', [TotalTerrainTriangles * 1.0]));
  MsgOut(Format('  Benchmark Results:', []));
  MsgOut(Format('    Total Time:      %.0n s', [BenchmarkLength]));
  MsgOut(Format('    Total Frames:    %.0n', [BenchmarkFrames * 1.0]));
  MsgOut(Format('    Total Triangles: %.0n', [BenchmarkTriangles * 1.0]));
  MsgOut(Format('    Average FPS: %4.0n', [AvgFPS]));
  MsgOut(Format('    Min FPS:     %4.0n', [BenchmarkMinFPS]));
  MsgOut(Format('    Max FPS:     %4.0n', [BenchmarkMaxFPS]));
  MsgOut(Format('    Average TPF: %7.0n', [AvgTPF]));
  MsgOut(Format('    Min TPF:     %7.0n', [BenchmarkMinTPF]));
  MsgOut(Format('    Max TPF:     %7.0n', [BenchmarkMaxTPF]));
  MsgOut(Format('    Average Render Time: %6.0n us', [BenchmarkRenderTime / BenchmarkFrames]));
  MsgOut(Format('    Average Refine Time: %6.0n us', [BenchmarkRefineTime / BenchmarkFrames]));
  MsgOut(Format('    Average Refinement Efficiency:  %3.3n%%', [AvgEfficiency]));
  MsgOut(Format('    Times Along Path: %s', [PathTimes]));
  MsgOut(Format('    FPS Along Path:   %s', [PathFPS]));
  MsgOut(Format('    TPS Along Path:   %s', [PathTPF]));
end;

procedure RunBenchmark(UpdateOnly: Boolean; const ElapsedTime: Single);
const
  WarmUpTime = 1.0;
var
  CurrFPS: Single;
begin
  if UpdateOnly then
  begin
    BenchmarkTime := BenchmarkTime + ElapsedTime;
    CurrFPS := 1.0 / ElapsedTime;

    if (BenchmarkStage = 0) or (BenchmarkTime > BenchmarkLength + WarmUpTime) then
    begin
      if BenchmarkStage > 0 then
      begin
        if CameraAnimation.FrameCount > 0 then
        begin
          CameraAnimation.Frames[CameraAnimation.FrameCount - 1].FPS := CurrFPS;
          CameraAnimation.Frames[CameraAnimation.FrameCount - 1].TPF := TrianglesPerFrame;
        end;
        StoreResults(BenchmarkStage);
      end;

      Inc(BenchmarkStage);

      if BenchmarkStage > BenchmarkStageCount  then
        EndBenchmark;

      // Setup various benchmarking scenarios
      case BenchmarkStage of
        1:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, 4 pixels)';
            TerrainSize := 1025;
            TerrainZScale := 0.25;
            TolerancePixels := 4;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        2:
          begin
            BenchmarkStageName := 'Mountains flyover (1 km, 4 pixels)';
            TerrainZScale := 0.25;
            HeightmapFile := GetAppDir + PathDelim + SMountainousHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SMountainousGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SMountainousPath, TerrainSize);
          end;
        3:
          begin
            BenchmarkStageName := 'Flatland flyover (1 km, 4 pixels)';
            TerrainZScale := 0.05;
            HeightmapFile := GetAppDir + PathDelim + SFlatHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SFlatGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SFlatPath, TerrainSize);
          end;
        4:
          begin
            BenchmarkStageName := 'Volcano flyover (0.5 km, 4 pixels)';
            TerrainSize := 513;
            TerrainZScale := 0.25;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        5:
          begin
            BenchmarkStageName := 'Volcano flyover (2 km, 4 pixels)';
            TerrainSize := 2049;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        6:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, 2 pixels)';
            TerrainSize := 1025;
            TolerancePixels := 2;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        7:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, 8 pixels)';
            TolerancePixels := 8;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        8:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, 4 pixels, isotropic error metric)';
            TolerancePixels := 4;
            ErrorMetric := emIsotropic;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        9:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, 4 pixels, culling off)';
            TolerancePixels := 4;
            ActiveCulling := False;
            ErrorMetric := emAnisotropic;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        10:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, full refinement)';
            ActiveCulling := True;
            ForceFullRefinement := True;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + SDefaultPath, TerrainSize);
          end;
        11:
          begin
            BenchmarkStageName := 'Volcano flyover (1 km, 4 pixels, 9x9 blocks)';
            ForceFullRefinement := False;
            TerrainBlocks := 9;
            HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
            GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
            LoadAnimation(CameraAnimation, GetAppDir + PathDelim + S9x9Path, TerrainSize);
          end;
      end;

      // Init app with current benchmark stage data
      FreeAppData;
      InitAppData;

      if CameraAnimation.FrameCount > 0 then
      begin
        Position := CameraAnimation.Frames[0].Position;
        Direction := CameraAnimation.Frames[0].Direction;
      end;

      // Reset benchmark results
      BenchmarkTime := 0.0;
      BenchmarkLength := CameraAnimation.Length;
      BenchmarkFrames := 0;
      BenchmarkTriangles := 0;
      BenchmarkMinFPS := 1e20;
      BenchmarkMaxFPS := 0.0;
      BenchmarkMinTPF := 1e20;
      BenchmarkMaxTPF := 0;
      BenchmarkRenderTime := 0;
      BenchmarkRefineTime := 0;
    end
    else
    begin
      // Continue current benchmarking stage

      // Wait a while before sampling of benchmark data starts
      if BenchmarkTime > WarmUpTime then
      begin
        // Update benchmark stats
        Inc(BenchmarkFrames);
        Inc(BenchmarkTriangles, TrianglesPerFrame);
        BenchmarkMinFPS := MinFloat(CurrFPS, BenchmarkMinFPS);
        BenchmarkMaxFPS := MaxFloat(CurrFPS, BenchmarkMaxFPS);
        BenchmarkMinTPF := MinFloat(TrianglesPerFrame, BenchmarkMinTPF);
        BenchmarkMaxTPF := MaxFloat(TrianglesPerFrame, BenchmarkMaxTPF);
        Inc(BenchmarkRenderTime, RenderTime);
        Inc(BenchmarkRefineTime, RefineTime);

        if AnimateCamera(CameraAnimation, BenchmarkTime - WarmUpTime, Position, Direction) and
          (CameraAnimation.CurrentFrame < CameraAnimation.FrameCount) then
        begin
          CameraAnimation.Frames[CameraAnimation.CurrentFrame].FPS := CurrFPS;
          CameraAnimation.Frames[CameraAnimation.CurrentFrame].TPF := TrianglesPerFrame;
        end;
      end;
    end;
  end
  else
  begin
    // Set default settings for benchmarking
    ScreenWidth := DefaultScreenWidth;
    ScreenHeight := DefaultScreenHeight;

    TerrainBlocks := 1;
    DrawFPS := False;
    ActiveWireOverlay := False;
    ActiveTexturing := True;
    ActiveFog := False;
    ForceFullRefinement := False;
    Anisotropy := False;
    DXTCompression := False;
    ActiveDetailTexture := False;
  end;
end;

procedure CheckInput(ElapsedTime: Single);
var
  AngleXY, AngleXZ, Speed: Single;
  X, Y: LongInt;
begin
{$IFDEF MSWINDOWS}
  // Do not process input when app window is not active
  if GetActiveWindow <> GetForeGroundWindow then
    Exit;
{$ENDIF}

  // Get angles between X/Y and X/Z planes from camera direction and update them
  // according to mouse movement
  AngleXY := ArcCos(Vec3Dot(Vec3Normalize(Vector3(Direction.X, Direction.Y, 0.0)), Vec3XUnit));
  if Direction.Y < 0.0 then
    AngleXY := - AngleXY;
  AngleXZ := -ArcSin(Direction.Z);
  AngleXY := AngleXY - ((MouseX - OldMouseX) * 0.01);
  AngleXZ := AngleXZ + ((MouseY - OldMouseY) * 0.01);
  AngleXZ := MinFloat(1.0, AngleXZ);
  AngleXZ := MaxFloat(-1.0, AngleXZ);

  ReturnMouse;

  Speed := WalkSpeed;
  if IsKeyDown(KEY_SHIFT) then
    Speed := RunSpeed;
  if IsKeyDown(KEY_CONTROL) then
    Speed := Speed * UltraSpeed;
  Speed := Speed * ElapsedTime;

  if IsKeyDown(KEY_UP) then
  begin
    Position.X := Position.X + Speed * Cos(AngleXY);
    Position.Y := Position.Y + Speed * Sin(AngleXY);
  end;
  if IsKeyDown(KEY_DOWN) then
  begin
    Position.X := Position.X - Speed * Cos(AngleXY);
    Position.Y := Position.Y - Speed * Sin(AngleXY);
  end;
  if IsKeyDown(KEY_LEFT) then
  begin
    Position.X := Position.X - Speed * Sin(AngleXY);
    Position.Y := Position.Y + Speed * Cos(AngleXY);
  end;
  if IsKeyDown(KEY_RIGHT) then
  begin
    Position.X := Position.X + Speed * Sin(AngleXY);
    Position.Y := Position.Y - Speed * Cos(AngleXY);
  end;
  if IsKeyDown(KEY_PAGEDOWN) then
    Position.Z := Position.Z - Speed;
  if IsKeyDown(KEY_PAGEUP) then
    Position.Z := Position.Z + Speed;

  if not ActiveFlying then
  begin
    X := Trunc(Position.X) div (TerrainSize - 1);
    Y := Trunc(Position.Y) div (TerrainSize - 1);
    Position.Z := GetTerrainZ(Position.X - X * (TerrainSize - 1),
      Position.Y - Y * (TerrainSize - 1), Position.Z);
  end;

  Direction := Vector3(Cos(AngleXZ) * Cos(AngleXY), Cos(AngleXZ) * Sin(AngleXY),
    -Sin(AngleXZ));
end;

procedure CheckCommandLine;
var
  I: LongInt;

  procedure ParseOption(const Opt: string);
  const
    STrue = 'true';
    SFalse = 'false';
  var
    I: LongInt;
    S, Arg: string;
  begin
    S := Opt;
    I := Pos('=', S);
    Arg := Copy(S, I + 1, MaxInt);
    Delete(S, I, MaxInt);
    S := LowerCase(S);
    if S = '-size' then
    begin
      if TryStrToInt(Arg, I) and (IsPow2(I) or IsPow2(I - 1)) then
        TerrainSize := I
      else
        WriteLn('Invalid terrain size: ', Arg, '. Using default value...');
    end
    else
    if S = '-blocks' then
    begin
      if TryStrToInt(Arg, I) then
        TerrainBlocks := I
      else
        WriteLn('Invalid terrain blocks: ', Arg, '. Using default value...');
    end
    else
    if S = '-scale' then
    begin
      if TryStrToInt(Arg, I) then
        TerrainZScale := I / 100
      else
        WriteLn('Invalid terrain Z scale: ', Arg, '. Using default value...');
    end
    else
    if S = '-startx' then
    begin
      if TryStrToInt(Arg, I) then
        StartX := I;
    end
    else
    if S = '-starty' then
    begin
      if TryStrToInt(Arg, I) then
        StartY := I;
    end
    else
    if S = '-map' then
      HeightmapFile := Arg
    else
    if S = '-texture' then
      GroundTexFile := Arg
    else
    if S = '-fullscreen' then
      FullScreen := SameText(Arg, STrue)
    else
    if S = '-anisotex' then
      Anisotropy := SameText(Arg, STrue)
    else
    if S = '-dxt' then
      DXTCompression := SameText(Arg, STrue)
    else
    if S = '-fog' then
      ActiveFog := SameText(Arg, STrue)
    else
    if S = '-detail' then
      ActiveDetailTexture := SameText(Arg, STrue)
    else
    if S = '-filter8bit' then
      Filter8BitHeightmaps := SameText(Arg, STrue)
    else
    if S = '-benchmark' then
      BenchmarkMode := SameText(Arg, STrue)
    else
    if S = '-anisometric' then
      ErrorMetric := TErrorMetric(Iff(SameText(Arg, STrue), Ord(emAnisotropic), Ord(emIsotropic)));
  end;

begin
  for I := 1 to ParamCount do
    ParseOption(ParamStr(I));
end;

{ GLUT callback procedures }

procedure Display; cdecl;
begin
  DrawScene;
  glutPostRedisplay;
end;

procedure Reshape(Width, Height: LongInt); cdecl;
begin
  ReSizeWindow(Width, Height);
  ReturnMouse;
end;

procedure Keyboard(Key: Byte; X, Y: LongInt); cdecl;
begin
  if Key = 27 then
    TerminateApp;

  if not BenchmarkMode then
  begin
    case Key of
      Ord('w'): ActiveWireOverlay := not ActiveWireOverlay;
      Ord('r'): ForceFullRefinement := not ForceFullRefinement;
      Ord('c'): ActiveCulling := not ActiveCulling;
      Ord('f'): ActiveFlying := not ActiveFlying;
      Ord('s'): ActiveStats := not ActiveStats;
      Ord('t'): ActiveTexturing := not ActiveTexturing;
      Ord('p'): TolerancePixels := Min(30, TolerancePixels + 1);
      Ord('o'): TolerancePixels := Max(1, TolerancePixels - 1);
      Ord('m'): MinLOD := Min(RefinementLevels, MinLOD + 1);
      Ord('n'): MinLOD := Max(0, MinLOD - 1);
      Ord('l'): CullFovX := MinFloat(90.0, CullFovX + 1.0);
      Ord('k'): CullFovX := MaxFloat(2.0, CullFovX - 1.0);
      Ord('d'): ActiveDetailTexture := not ActiveDetailTexture;
      Ord('e'): ErrorMetric := TErrorMetric(Iff(ErrorMetric <> emAnisotropic,
        Ord(emAnisotropic), Ord(emIsotropic)));
      Ord('g'):
        begin
          ActiveFog := not ActiveFog;
          if ActiveFog then
            glEnable(GL_FOG)
          else
            glDisable(GL_FOG);
        end;
    end;
  end;
end;

procedure Special(Key, X, Y: LongInt); cdecl;
begin
end;

procedure PassiveMotion(X, Y: LongInt); cdecl;
begin
  MouseX := X;
  MouseY := Y;
end;

procedure Menu(Value: LongInt); cdecl;
begin
  if Value < 0 then
    Special(-Value, -1, -1)
  else
    Keyboard(Value, -1, -1);
  ReturnMouse;
end;

{ Program entry point }

var
  Count, Render, LOD, Other: LongInt;
  Param: PChar;
begin
  SysUtils.DecimalSeparator := '.';
  SysUtils.ThousandSeparator := ',';

  HeightmapFile := GetAppDir + PathDelim + SDefaultHeightmap;
  GroundTexFile := GetAppDir + PathDelim + SDefaultGround;
  TerrainSize := DefaultTerrainSize;

  MsgOut('glSOAR ' + STimeStamp);
  MsgOut('by Marek Mauder');

  // Check if GL, GLU, and GLUT libraries are loaded
  if @glLoadIdentity = nil then
  begin
    MsgOut('GL library not loaded successfully');
    TerminateApp;
  end;
  if @gluPerspective = nil then
  begin
    MsgOut('GLU library not loaded successfully');
    TerminateApp;
  end;
  if @glutInit = nil then
  begin
    MsgOut('GLUT library not loaded successfully');
    TerminateApp;
  end;

  CheckCommandLine;
  Position := Vector3(StartX / 100 * TerrainSize, StartY / 100 * TerrainSize, 0);

{$IF defined(LINUX) or defined(UNIX)}
  if not BenchmarkMode then
  begin
    MsgOut('Only Benchmark Mode allowed in Linux/Unix (glSOAR -benchmark=true)');
    TerminateApp;
  end;
{$IFEND}

  Count := 1;
  Param := PChar(ParamStr(0));
  try
    if BenchmarkMode then
      RunBenchmark(False, 0);

    glutInit(@Count, @Param);
    glutInitDisplayMode(GLUT_DOUBLE or GLUT_RGBA or GLUT_DEPTH);
    glutInitWindowSize(ScreenWidth, ScreenHeight);
    glutInitWindowPosition((glutGet(GLUT_SCREEN_WIDTH) - ScreenWidth) div 2,
      (glutGet(GLUT_SCREEN_HEIGHT) - ScreenHeight) div 2);
    glutCreateWindow(SWinTitle);
    if FullScreen then
      glutFullScreen;

    glutReshapeFunc(Reshape);
    glutDisplayFunc(Display);
    glutKeyboardFunc(Keyboard);
    glutSpecialFunc(Special);
    glutPassiveMotionFunc(PassiveMotion);

    if BenchmarkMode then
      RunBenchmark(True, 0)
    else
      InitAppData;

    Render := glutCreateMenu(Menu);
    glutAddMenuEntry('t - Toggle Texturing', Ord('t'));
    glutAddMenuEntry('w - Toggle Wireframe Overlay', Ord('w'));
    glutAddMenuEntry('d - Detail Texture', Ord('d'));
    glutAddMenuEntry('g - Toggle Fog', Ord('g'));
    Lod := glutCreateMenu(Menu);
    glutAddMenuEntry('p - Increase Tolerance (User error)', Ord('p'));
    glutAddMenuEntry('o - Decrease Tolerance (User error)', Ord('o'));
    glutAddMenuEntry('l - Increase Culling FOV', Ord('l'));
    glutAddMenuEntry('k - Decrease Culling FOV', Ord('k'));
    glutAddMenuEntry('m - Increase Refinement Min LOD', Ord('m'));
    glutAddMenuEntry('n - Decrease Refinement Min LOD', Ord('n'));
    glutAddMenuEntry('c - Toggle View Culling', Ord('c'));
    glutAddMenuEntry('r - Toggle Full Refinement', Ord('r'));
    glutAddMenuEntry('e - Switch Error Metric', Ord('e'));
    Other := glutCreateMenu(Menu);
    glutAddMenuEntry('s - Toggle Stats & Info', Ord('s'));
    glutAddMenuEntry('f - Toggle Flying', Ord('f'));
    glutCreateMenu(Menu);
    glutAddSubMenu('Rendering', Render);
    glutAddSubMenu('SOAR', LOD);
    glutAddSubMenu('Other', Other);
    glutAddMenuEntry('ESC - Exit', 27);

    if not BenchmarkMode then
      glutAttachMenu(GLUT_RIGHT_BUTTON);

    glutMainLoop;
  except
    on E: Exception do
    begin
      MsgOut('Exception raised!');
      MsgOut('Message: ' + E.Message);
    end;
  end;
  TerminateApp;
end.
